# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ui/select_list.ui'
#
# Created by: PyQt5 UI code generator 5.5.1
#
# WARNING! All changes made in this file will be lost!

from ..qt import QtCore, QtGui, QtWidgets

class Ui_SelectListDialog(object):
    def setupUi(self, SelectListDialog):
        SelectListDialog.setObjectName("SelectListDialog")
        SelectListDialog.resize(335, 300)
        self.verticalLayout = QtWidgets.QVBoxLayout(SelectListDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.label = QtWidgets.QLabel(SelectListDialog)
        self.label.setWordWrap(True)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        self.listWidget = QtWidgets.QListWidget(SelectListDialog)
        self.listWidget.setObjectName("listWidget")
        self.verticalLayout.addWidget(self.listWidget)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.okButton = QtWidgets.QPushButton(SelectListDialog)
        self.okButton.setObjectName("okButton")
        self.horizontalLayout.addWidget(self.okButton)
        self.cancelButton = QtWidgets.QPushButton(SelectListDialog)
        self.cancelButton.setObjectName("cancelButton")
        self.horizontalLayout.addWidget(self.cancelButton)
        self.verticalLayout.addLayout(self.horizontalLayout)

        self.retranslateUi(SelectListDialog)
        QtCore.QMetaObject.connectSlotsByName(SelectListDialog)

    def retranslateUi(self, SelectListDialog):
        _translate = QtCore.QCoreApplication.translate
        SelectListDialog.setWindowTitle(_translate("SelectListDialog", "Select your fs2_open_* executable."))
        self.label.setText(_translate("SelectListDialog", "Please select the fs2_open_* file you want to use."))
        self.okButton.setText(_translate("SelectListDialog", "OK"))
        self.cancelButton.setText(_translate("SelectListDialog", "Cancel"))




## Copyright 2017 Knossos authors, see NOTICE file
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from __future__ import absolute_import, print_function

import os
import sys
import logging
import shlex
import functools
import json
import semantic_version

from . import uhf
uhf(__name__)

from . import center, util, integration, api, web, repo
from .qt import QtCore, QtGui, QtWidgets, load_styles
from .ui.hell import Ui_MainWindow as Ui_Hell
from .ui.gogextract import Ui_GogExtractDialog
from .ui.flags import Ui_FlagsDialog
from .ui.install import Ui_InstallDialog
from .ui.mod_settings import Ui_ModSettingsDialog
from .ui.mod_versions import Ui_ModVersionsDialog
from .ui.log_viewer import Ui_LogDialog
from .tasks import run_task, GOGExtractTask, InstallTask, UninstallTask, WindowsUpdateTask, CheckFilesTask

# Keep references to all open windows to prevent the GC from deleting them.
_open_wins = []
translate = QtCore.QCoreApplication.translate


class QDialog(QtWidgets.QDialog):

    def __init__(self, *args):
        super(QDialog, self).__init__(*args)
        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)


class QMainWindow(QtWidgets.QMainWindow):

    def __init__(self, *args):
        super(QMainWindow, self).__init__(*args)
        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)

    def closeEvent(self, e):
        e.accept()

        center.app.quit()

    def changeEvent(self, event):
        if event.type() == QtCore.QEvent.ActivationChange and integration.current is not None:
            integration.current.annoy_user(False)

        return super(QMainWindow, self).changeEvent(event)


class Window(object):
    win = None
    closed = True
    _is_window = True
    _tpl_cache = None
    _styles = ''
    _qchildren = None

    def __init__(self, window=True):
        super(Window, self).__init__()

        self._is_window = window
        self._tpl_cache = {}
        self._qchildren = []

    def _create_win(self, ui_class, qt_widget=QDialog):
        if not self._is_window:
            qt_widget = QtWidgets.QWidget

        self.win = util.init_ui(ui_class(), qt_widget())

    def open(self):
        global _open_wins

        if not self.closed:
            return

        self.closed = False
        _open_wins.append(self)
        self.win.destroyed.connect(self._del)
        self.win.show()

    def close(self):
        self.win.close()

    def _del(self):
        global _open_wins

        if self.closed:
            return

        self.closed = True

        if self in _open_wins:
            _open_wins.remove(self)

    def label_tpl(self, label, **vars):
        name = label.objectName()
        if name in self._tpl_cache:
            text = self._tpl_cache[name]
        else:
            text = self._tpl_cache[name] = label.text()

        for name, value in vars.items():
            text = text.replace('{' + name + '}', value)

        label.setText(text)

    def load_styles(self, path, append=True):
        data = load_styles(path)

        if append:
            self._styles += data
        else:
            self._styles = data

        self.win.setStyleSheet(self._styles)

    def tr(self, *args):
        return translate(self.__class__.__name__, *args)


class HellWindow(Window):
    _tasks = None
    _mod_filter = 'home'
    _search_text = ''
    _updating_mods = None
    browser_ctrl = None
    progress_win = None

    def __init__(self, window=True):
        super(HellWindow, self).__init__(window)
        self._tasks = {}
        self._updating_mods = {}

        self._create_win(Ui_Hell, QMainWindow)
        self.browser_ctrl = web.BrowserCtrl(self.win.webView)
        self.win.verticalLayout.setContentsMargins(0, 0, 0, 0)

        self.win.webView.loadStarted.connect(self.show_indicator)
        self.win.webView.loadFinished.connect(self.check_loaded)

        center.signals.update_avail.connect(self.ask_update)
        center.signals.task_launched.connect(self.watch_task)
        center.signals.repo_updated.connect(self.update_mod_list)

        self.win.setWindowTitle(self.win.windowTitle() + ' ' + center.VERSION)
        self.win.progressInfo.hide()
        self.open()

    def _del(self):
        center.signals.update_avail.disconnect(self.ask_update)
        center.signals.task_launched.disconnect(self.watch_task)

        # Trying to free this object usually leads to memory corruption
        # See http://pyqt.sourceforge.net/Docs/PyQt5/gotchas.html#crashes-on-exit
        # Since this is the main window and should only be closed whenever the application exits, skipping this
        # won't lead to memory leaks.
        # super(HellWindow, self)._del()

    def finish_init(self):
        self.check_fso()
        api.init_self()

    def check_fso(self):
        if 'KN_WELCOME' not in os.environ and center.settings['base_path'] is not None:
            self.update_mod_buttons('home')
        else:
            self.browser_ctrl.bridge.showWelcome.emit()

    def ask_update(self, version):
        # We only have an updater for windows.
        if sys.platform.startswith('win'):
            msg = self.tr('There\'s an update available!\nDo you want to install Knossos %s now?') % str(version)
            buttons = QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No
            result = QtWidgets.QMessageBox.question(center.app.activeWindow(), 'Knossos', msg, buttons)
            if result == QtWidgets.QMessageBox.Yes:
                run_task(WindowsUpdateTask())
        else:
            msg = self.tr('There\'s an update available!\nYou should update to Knossos %s.') % str(version)
            QtWidgets.QMessageBox.information(center.app.activeWindow(), 'Knossos', msg)

    def update_repo_list(self):
        api.fetch_list()

    def search_mods(self):
        mods = None

        if self._mod_filter in ('home', 'develop'):
            mods = center.installed.mods
        elif self._mod_filter == 'explore':
            mods = {}
            for mid, mvs in center.mods.mods.items():
                if mid not in center.installed.mods:
                    mods[mid] = mvs
        # elif self._mod_filter == 'updates':
        #     mods = {}
        #     for mid in center.installed.get_updates():
        #         mods[mid] = center.installed.mods[mid]
        else:
            mods = {}

        # Now filter the mods.
        query = self._search_text
        result = []
        for mid, mvs in mods.items():
            if query in mvs[0].title.lower():
                mod = mvs[0]
                item = mod.get()
                item['progress'] = 0
                item['progress_info'] = {}

                try:
                    rmod = center.mods.query(mod)
                except repo.ModNotFound:
                    rmod = None

                if rmod and rmod.version > mod.version:
                    item['status'] = 'update'
                elif mod.mid in self._updating_mods:
                    item['status'] = 'updating'
                    item['progress'] = self._updating_mods[mod.mid]
                else:
                    item['status'] = 'ready'

                    if self._mod_filter == 'home':
                        for pkg in mod.packages:
                            if pkg.files_checked > 0 and pkg.files_ok < pkg.files_checked:
                                item['status'] = 'error'
                                break

                result.append(item)

        result.sort(key=lambda m: m['title'])
        return result, self._mod_filter

    def update_mod_list(self):
        result, filter_ = self.search_mods()

        if filter_ in ('home', 'explore', 'develop'):
            self.browser_ctrl.bridge.updateModlist.emit(result, filter_)

    def show_settings(self):
        pass

    def show_indicator(self):
        self.win.setCursor(QtCore.Qt.BusyCursor)

    def check_loaded(self, success):
        self.win.unsetCursor()

    def update_mod_buttons(self, clicked=None):
        if center.settings['base_path'] is not None:
            self._mod_filter = clicked
            self.update_mod_list()

    def perform_search(self, term):
        self._search_text = term.lower()
        self.update_mod_list()

    def watch_task(self, task):
        logging.debug('Task "%s" (%d, %s) started.', task.title, id(task), task.__class__)
        self._tasks[id(task)] = task
        self.browser_ctrl.bridge.taskStarted.emit(id(task), task.title, [m.mid for m in task.mods])

        for m in task.mods:
            self._updating_mods[m.mid] = 0

        task.done.connect(functools.partial(self._forget_task, task))
        task.progress.connect(functools.partial(self._track_progress, task))

        if len(task.mods) == 0:
            if len(self._tasks) == 1:
                self.win.progressInfo.show()
                self.win.progressLabel.setText(task.title)
                self.win.progressBar.setValue(0)

                integration.current.show_progress(0)
            else:
                # TODO: Stop being lazy and calculate the aggregate progress.
                self.win.progressBar.hide()
                self.win.progressLabel.setText(self.tr('Working...'))

    def _track_progress(self, task, pi):
        self.browser_ctrl.bridge.taskProgress.emit(id(task), pi[0] * 100, json.dumps(pi[1]))

        for m in task.mods:
            self._updating_mods[m.mid] = pi[0] * 100

        if len(self._tasks) == 1:
            integration.current.set_progress(pi[0])
            self.win.progressBar.setValue(pi[0] * 100)

    def _forget_task(self, task):
        logging.debug('Task "%s" (%d) finished.', task.title, id(task))
        self.browser_ctrl.bridge.taskFinished.emit(id(task))
        del self._tasks[id(task)]

        for m in task.mods:
            if m.mid in self._updating_mods:
                del self._updating_mods[m.mid]

        if len(self._tasks) == 1:
            task = list(self._tasks.values())[0]
            self.win.progressLabel.setText(task.title)
            self.win.progressBar.setValue(task.get_progress()[0])
            self.win.progressBar.show()
        elif len(self._tasks) == 0:
            self.win.progressInfo.hide()
            integration.current.hide_progress()

    def abort_task(self, task):
        if task in self._tasks:
            self._tasks[task].abort()


class GogExtractWindow(Window):

    def __init__(self, window=True):
        super(GogExtractWindow, self).__init__(window)

        self._create_win(Ui_GogExtractDialog)

        self.win.gogPath.textChanged.connect(self.validate)
        self.win.destPath.textChanged.connect(self.validate)

        self.win.gogButton.clicked.connect(self.select_installer)
        self.win.destButton.clicked.connect(self.select_dest)
        self.win.cancelButton.clicked.connect(self.win.close)
        self.win.installButton.clicked.connect(self.do_install)

        if window:
            self.win.setModal(True)
            self.open()

    def select_installer(self):
        path = QtWidgets.QFileDialog.getOpenFileName(self.win,
            self.tr('Please select the setup_freespace2_*.exe file.'),
            os.path.expanduser('~/Downloads'), self.tr('Executable (*.exe)'))

        if isinstance(path, tuple):
            path = path[0]

        if path is not None and path != '':
            if not os.path.isfile(path):
                QtWidgets.QMessageBox.critical(self.win, self.tr('Not a file'), self.tr('Please select a proper file!'))
                return

            self.win.gogPath.setText(os.path.abspath(path))

    def select_dest(self):
        path = QtWidgets.QFileDialog.getExistingDirectory(self.win, self.tr('Please select the destination directory.'),
            os.path.expanduser('~/Documents'))

        if path is not None and path != '':
            if not os.path.isdir(path):
                QtWidgets.QMessageBox.critical(self.win, self.tr('Not a directory'),
                    self.tr('Please select a proper directory!'))
                return

            self.win.destPath.setText(os.path.abspath(path))

    def validate(self):
        if os.path.isfile(self.win.gogPath.text()) and os.path.isdir(self.win.destPath.text()):
            self.win.installButton.setEnabled(True)
        else:
            self.win.installButton.setEnabled(False)

    def do_install(self):
        # Just to be sure...
        if os.path.isfile(self.win.gogPath.text()) and os.path.isdir(self.win.destPath.text()):
            center.main_win.update_mod_buttons('progress')

            run_task(GOGExtractTask(self.win.gogPath.text(), self.win.destPath.text()))

            if self._is_window:
                self.close()


class FlagsWindow(Window):
    _flags = None
    _selected = None
    _mod = None
    _mod_arg = ''

    def __init__(self, mod=None, window=True):
        super(FlagsWindow, self).__init__(window)

        self._selected = []
        self._mod = mod

        self._create_win(Ui_FlagsDialog)
        if window:
            self.win.setModal(True)

            self.win.okButton.clicked.connect(self.win.accept)
            self.win.cancelButton.clicked.connect(self.win.reject)
            self.win.accepted.connect(self.save)
            self.win.saveButton.hide()
        else:
            self.win.saveButton.clicked.connect(self.save)
            self.win.okButton.hide()
            self.win.cancelButton.hide()

        self.win.easySetup.activated.connect(self._set_easy)
        self.win.easySetup.activated.connect(self.update_display)
        self.win.listType.activated.connect(self._update_list)
        self.win.customFlags.textEdited.connect(self.update_display)
        self.win.flagList.itemClicked.connect(self.update_display)

        self.win.defaultsButton.clicked.connect(self.set_defaults)

        if window:
            self.open()

        if mod is None:
            self.win.defaultsButton.hide()
        else:
            try:
                self._mod_arg = ','.join(mod.get_mod_flag())
            except repo.ModNotFound:
                pass

        self.read_flags()

    def read_flags(self):
        flags = api.get_fso_flags()

        if flags is None:
            self.win.setEnabled(False)
            self.win.cmdLine.setPlainText(self.tr('Until you select a working FS2 build, I won\'t be able to help you.'))
            return

        self.win.setEnabled(True)
        self.win.easySetup.clear()
        self.win.listType.clear()

        for key, name in flags.easy_flags.items():
            self.win.easySetup.addItem(name, key)

        self._flags = flags.flags
        for name in self._flags:
            self.win.listType.addItem(name)

        self.set_selection(api.get_cmdline(self._mod))

    def _set_easy(self):
        if self._flags is None:
            return

        combo = self.win.easySetup
        cur_mode = combo.itemData(combo.currentIndex())

        # Update self._selected.
        self.get_selection()

        for flags in self._flags.values():
            for info in flags:
                if info['on_flags'] & cur_mode == cur_mode and info['name'] not in self._selected:
                    self._selected.append(info['name'])

                if info['off_flags'] & cur_mode == cur_mode and info['name'] in self._selected:
                    self._selected.remove(info['name'])

        self._update_list(save_selection=False)

    def _update_list(self, idx=None, save_selection=True):
        if self._flags is None:
            return

        cur_type = self.win.listType.currentText()

        if save_selection:
            # Remember the currently selected options.
            self.get_selection()

        self.win.flagList.clear()
        for flag in self._flags[cur_type]:
            label = flag['desc']
            if label == '':
                label = flag['name']

            item = QtWidgets.QListWidgetItem(label)
            item.setData(QtCore.Qt.UserRole, flag['name'])
            if flag['name'] in self._selected:
                item.setCheckState(QtCore.Qt.Checked)
            else:
                item.setCheckState(QtCore.Qt.Unchecked)

            self.win.flagList.addItem(item)

    def get_selection(self, split_mod_flag=False):
        count = self.win.flagList.count()

        for i in range(count):
            item = self.win.flagList.item(i)
            flag = item.data(QtCore.Qt.UserRole)
            if item.checkState() == QtCore.Qt.Checked:
                if flag not in self._selected:
                    self._selected.append(flag)
            else:
                if flag in self._selected:
                    self._selected.remove(flag)

        try:
            custom = shlex.split(self.win.customFlags.text())
        except ValueError:
            # Invalid user input...
            custom = []

        if split_mod_flag:
            idx = -1
            for i, flag in enumerate(custom):
                if flag == '-mod':
                    idx = i
                    break

            if idx > -1 and idx + 1 < len(custom):
                mod_arg = ['-mod', custom[idx + 1]]
                custom = custom[:idx] + custom[idx + 2:]
            else:
                mod_arg = []

            return self._selected + custom, mod_arg
        else:
            return self._selected + custom

    def update_display(self):
        if self.mod:
            fs2_bin = self.mod.get_bin()
        else:
            fs2_bin = center.settings['fs2_bin']

        sel, mod_flag = self.get_selection(True)
        if self._mod is not None:
            if len(mod_flag) > 0:
                mod_flag[1] = mod_flag[1].strip(',') + ',' + self._mod_arg
            else:
                mod_flag = ['-mod', self._mod_arg]

        cmdline = ' '.join([fs2_bin] + [shlex.quote(opt) for opt in sel + mod_flag])

        self.win.cmdLine.setPlainText(cmdline)

    def set_selection(self, flags):
        self._selected = flags[:]
        custom = []
        all_flags = []

        for flags in self._flags.values():
            for info in flags:
                all_flags.append(info['name'])

        for opt in self._selected[:]:
            if opt not in all_flags:
                custom.append(shlex.quote(opt))
                self._selected.remove(opt)

        self.win.customFlags.setText(' '.join(custom))
        self._update_list(save_selection=False)
        self.update_display()

    def set_defaults(self):
        cmdlines = center.settings['cmdlines']

        if '#default' in cmdlines:
            self.set_selection(cmdlines['#default'])
        else:
            self.set_selection([])

        self.save()

    def save(self):
        sel = self.get_selection()
        clines = center.settings['cmdlines']

        if self._mod is None:
            clines['#default'] = self.get_selection()
        else:
            mid = self._mod.mid
            if sel == clines['#default']:
                logging.info('Not saving default flags!')
                if mid in clines:
                    del clines[mid]
            else:
                clines[mid] = self.get_selection()

        api.save_settings()


class ModInstallWindow(Window):
    _mod = None
    _pkg_checks = None
    _dep_tpl = None

    def __init__(self, mod, sel_pkgs=[]):
        super(ModInstallWindow, self).__init__()

        self._mod = mod
        self._create_win(Ui_InstallDialog)
        self.win.splitter.setStretchFactor(0, 2)
        self.win.splitter.setStretchFactor(1, 1)

        self.label_tpl(self.win.titleLabel, MOD=mod.title)
        self.win.notesField.setPlainText(mod.notes)

        self._pkg_checks = []
        self.show_packages()
        self.update_deps()

        self.win.treeWidget.itemChanged.connect(self.update_selection)

        self.win.accepted.connect(self.install)
        self.win.rejected.connect(self.close)

        self.open()

    def show_packages(self):
        pkgs = self._mod.packages
        try:
            all_pkgs = center.mods.process_pkg_selection(pkgs)
        except repo.ModNotFound as exc:
            logging.exception('Well, I won\'t be installing that...')
            msg = self.tr('I\'m sorry but you won\'t be able to install "%s" because "%s" is missing!') % (
                self._mod.title, exc.mid)
            QtWidgets.QMessageBox.critical(None, 'Knossos', msg)

            self.close()
            return

        needed_pkgs = self._mod.resolve_deps()
        mods = set()

        for pkg in all_pkgs:
            mods.add(pkg.get_mod())

        # Make sure our current mod comes first.
        if self._mod in mods:
            mods.remove(self._mod)

        mods = [self._mod] + list(mods)
        for mod in mods:
            item = QtWidgets.QTreeWidgetItem(self.win.treeWidget, [mod.title, ''])
            item.setExpanded(True)
            item.setData(0, QtCore.Qt.UserRole, mod)
            c = 0

            for pkg in mod.packages:
                fsize = 0
                for f_item in pkg.files.values():
                    fsize += f_item.get('filesize', 0)

                if fsize > 0:
                    fsize = util.format_bytes(fsize)
                else:
                    fsize = '?'

                sub = QtWidgets.QTreeWidgetItem(item, [pkg.name, fsize])
                is_installed = center.installed.is_installed(pkg)

                if is_installed:
                    sub.setText(1, self.tr('Installed'))

                if pkg.status == 'required' or pkg in needed_pkgs or is_installed:
                    sub.setCheckState(0, QtCore.Qt.Checked)
                    sub.setDisabled(True)

                    c += 1
                elif pkg.status == 'recommended':
                    sub.setCheckState(0, QtCore.Qt.Checked)
                    c += 1
                if pkg.status == 'optional':
                    sub.setCheckState(0, QtCore.Qt.Unchecked)

                sub.setData(0, QtCore.Qt.UserRole, pkg)
                sub.setData(0, QtCore.Qt.UserRole + 1, is_installed)
                self._pkg_checks.append(sub)

            if c == len(mod.packages):
                item.setCheckState(0, QtCore.Qt.Checked)
            elif c > 0:
                item.setCheckState(0, QtCore.Qt.PartiallyChecked)
            else:
                item.setCheckState(0, QtCore.Qt.Unchecked)

        self.win.treeWidget.header().resizeSections(QtWidgets.QHeaderView.ResizeToContents)

    def update_selection(self, item, col):
        obj = item.data(0, QtCore.Qt.UserRole)
        if isinstance(obj, repo.Mod):
            cs = item.checkState(0)

            for i in range(item.childCount()):
                child = item.child(i)
                if not child.isDisabled():
                    child.setCheckState(0, cs)

        self.update_deps()

    def get_selected_pkgs(self):
        pkgs = []
        for item in self._pkg_checks:
            if item.checkState(0) == QtCore.Qt.Checked:
                pkgs.append(item.data(0, QtCore.Qt.UserRole))

        pkgs = center.mods.process_pkg_selection(pkgs)
        return pkgs

    def update_deps(self):
        pkg_sel = self.get_selected_pkgs()
        dl_size = 0
        for item in self._pkg_checks:
            pkg = item.data(0, QtCore.Qt.UserRole)

            if not item.data(0, QtCore.Qt.UserRole + 1):
                if item.checkState(0) == QtCore.Qt.Checked or pkg in pkg_sel:
                    for f_item in pkg.files.values():
                        dl_size += f_item.get('filesize', 0)

            if item.isDisabled():
                continue

            if pkg in pkg_sel:
                item.setCheckState(0, QtCore.Qt.Checked)

        self.label_tpl(self.win.dlSizeLabel, DL_SIZE=util.format_bytes(dl_size))

    def install(self):
        center.main_win.update_mod_buttons('home')

        run_task(InstallTask(self.get_selected_pkgs(), self._mod))
        self.close()


class ModSettingsWindow(Window):
    _mod = None
    _pkg_checks = None
    _mod_versions = None
    _flags = None

    def __init__(self, mod, window=True):
        super(ModSettingsWindow, self).__init__(window)

        self._mod = mod
        self._mod_versions = []
        self._create_win(Ui_ModSettingsDialog, QDialog)
        self.load_styles(':/ui/themes/default/mod_settings.css')

        self.win.modTitle.setText(self._mod.title)
        self.win.modLogo.setPixmap(QtGui.QPixmap(mod.logo_path))
        self.win.modDesc.setPlainText(mod.description)

        lay = QtWidgets.QVBoxLayout(self.win.flagsTab)
        self._flags = FlagsWindow(mod, False)
        lay.addWidget(self._flags.win)

        self._pkg_checks = []
        if isinstance(mod, repo.IniMod):
            self.win.tabWidget.setTabEnabled(2, False)  # Packages
            self.win.tabWidget.setTabEnabled(3, False)  # Versions
            self.win.tabWidget.setTabEnabled(4, False)  # Troubleshooting
        else:
            try:
                rmod = center.mods.query(mod)
            except repo.ModNotFound:
                rmod = mod

            for pkg in rmod.packages:
                p_check = QtWidgets.QCheckBox(pkg.name)
                installed = center.installed.is_installed(pkg)

                if pkg.status == 'required' or installed:
                    p_check.setCheckState(QtCore.Qt.Checked)

                    if pkg.status == 'required':
                        p_check.setDisabled(True)

                        if not installed:
                            p_check.setProperty('error', True)
                            p_check.setText(pkg.name + self.tr(' (missing)'))
                else:
                    p_check.setCheckState(QtCore.Qt.Unchecked)

                p_check.stateChanged.connect(self.update_dlsize)
                self.win.pkgsLayout.addWidget(p_check)
                self._pkg_checks.append([p_check, installed, pkg])

            settings = center.settings['mod_settings'].get(self._mod.mid)
            if settings:
                if settings.get('parse_mod_ini', False):
                    self.win.parseModIni.setCheckState(QtCore.Qt.Checked)
                else:
                    self.win.parseModIni.setCheckState(QtCore.Qt.Unchecked)

            self.win.applyPkgChanges.clicked.connect(self.apply_pkg_selection)

            self.win.parseModIni.stateChanged.connect(self.apply_mod_ini_setting)
            self.win.checkFiles.clicked.connect(self.check_files)
            self.win.delLoose.clicked.connect(self.delete_loose_files)
            self.win.replaceFiles.clicked.connect(self.repair_files)

            self.update_dlsize()
            self.show_versions()

        center.signals.repo_updated.connect(self.update_repo_related)

        self.show_flags_tab()
        self.open()

    def _del(self):
        center.signals.repo_updated.disconnect(self.update_repo_related)

        super(ModSettingsWindow, self)._del()

    def update_repo_related(self):
        try:
            self._mod = center.installed.query(self._mod)
        except repo.ModNotFound:
            self.close()

            # Try to open a new window showing the latest version.
            try:
                m2 = center.installed.query(self._mod.mid)
            except repo.ModNotFound:
                pass
            else:
                ModSettingsWindow(m2)

            return

        self.show_versions()

        for i, item in enumerate(self._pkg_checks):
            checked = item[0].checkState() == QtCore.Qt.Checked
            was_inst = item[1]
            is_inst = center.installed.is_installed(item[2])

            if was_inst != is_inst:
                if checked == was_inst:
                    if is_inst:
                        item[0].setCheckState(QtCore.Qt.Checked)
                    else:
                        item[0].setCheckState(QtCore.Qt.Unchecked)

                if is_inst and item[0].property('error'):
                    item[0].setProperty('error', False)
                    item[0].setText(item[2].name)
                elif not is_inst and item[2].status == 'required' and not item[0].property('error'):
                    item[0].setProperty('error', True)
                    item[0].setText(item[2].name + self.tr(' (missing)'))

                item[1] = is_inst

        self.update_dlsize()

    def show_flags_tab(self):
        self.win.tabWidget.setCurrentIndex(0)

    def show_pkg_tab(self):
        self.win.tabWidget.setCurrentIndex(1)

    def get_selected_pkgs(self):
        install = []
        remove = []

        for check, is_installed, pkg in self._pkg_checks:
            selected = check.checkState() == QtCore.Qt.Checked
            if selected != is_installed:
                if selected:
                    install.append(pkg)
                else:
                    remove.append(pkg)

        if len(install) > 0:
            try:
                install = center.mods.process_pkg_selection(install)
            except repo.ModNotFound as exc:
                if center.mods.has(self._mod):
                    QtWidgets.QMessageBox.critical(None, 'Knossos',
                        self.tr("I'm sorry but I can't install the selected packages because the dependency " +
                            '"%s" is missing!') % exc.mid)
                else:
                    QtWidgets.QMessageBox.critical(None, 'Knossos', self.tr("I'm sorry but I can't install new " +
                        "packages for this mod since it's not available anymore!"))

                install = []

        remove = list(set(remove) - set(install))

        return install, remove

    def update_dlsize(self):
        install, remove = self.get_selected_pkgs()
        try:
            all_pkgs = center.mods.process_pkg_selection(install)
        except repo.ModNotFound:
            all_pkgs = []

        dl_size = 0

        for pkg in all_pkgs:
            if not center.installed.is_installed(pkg):
                for item in pkg.files.values():
                    dl_size += item.get('filesize', 0)

        self.label_tpl(self.win.dlSizeLabel, DL_SIZE=util.format_bytes(dl_size))

    def apply_pkg_selection(self):
        install, remove = self.get_selected_pkgs()

        if len(remove) == 0 and len(install) == 0:
            # Nothing to do...
            return

        msg = ''
        if len(remove) > 0:
            msg += self.tr("I'm going to remove %s.\n") % util.human_list([p.name for p in remove])

        if len(install) > 0:
            msg += self.tr("I'm going to install %s.\n") % util.human_list(
                [p.name for p in install if not center.installed.is_installed(p)])

        box = QtWidgets.QMessageBox()
        box.setIcon(QtWidgets.QMessageBox.Question)
        box.setText(msg)
        box.setInformativeText(self.tr('Continue?'))
        box.setStandardButtons(QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
        box.setDefaultButton(QtWidgets.QMessageBox.Yes)

        if box.exec_() == QtWidgets.QMessageBox.Yes:
            run_task(UninstallTask(remove))
            run_task(InstallTask(install))

    def show_versions(self):
        mods = set()
        try:
            for dep in self._mod.resolve_deps():
                mods.add(dep.get_mod())
        except repo.ModNotFound as exc:
            QtWidgets.QMessageBox.critical(None, 'Knossos',
                self.tr('This mod is missing the dependency "%s"!') % exc.mid)
            return

        mods.add(self._mod)

        layout = self.win.versionsTab.layout()
        if layout:
            # Clear the current layout
            item = layout.takeAt(0)
            while item:
                w = item.widget()
                if w:
                    w.deleteLater()

                item = layout.takeAt(0)
        else:
            layout = QtWidgets.QGridLayout()
            self.win.versionsTab.setLayout(layout)

        mods = sorted(mods, key=lambda m: m.title)
        for i, mod in enumerate(mods):
            versions = list(center.installed.query_all(mod.mid))
            label = QtWidgets.QLabel(mod.title + ': ')
            label.setWordWrap(True)

            sel = QtWidgets.QComboBox()
            sel.addItem(self.tr('Latest (%s)') % versions[0].version, None)

            pin = center.installed.get_pin(mod)
            for n, mv in enumerate(versions):
                sel.addItem(str(mv.version), mv.version)

                if pin == mv.version:
                    sel.setCurrentIndex(n + 1)

            editBut = QtWidgets.QPushButton(self.tr('Edit'))

            layout.addWidget(label, i, 0)
            layout.addWidget(sel, i, 1)
            layout.addWidget(editBut, i, 2)

            if mod == self._mod:
                cb_sel = functools.partial(self.update_versions, sel, mod)
                cb_edit = functools.partial(self.open_ver_edit, mod)

                sel.currentIndexChanged.connect(cb_sel)
                editBut.clicked.connect(cb_edit)
            else:
                # TODO: Finish code for pinning dependencies.
                sel.setEnabled(False)
                editBut.setEnabled(False)

        layout.addItem(QtWidgets.QSpacerItem(0, 0, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding))

    def update_versions(self, sel, mod, idx):
        version = sel.itemData(idx)

        if mod == self._mod:
            if version is None:
                center.installed.unpin(self._mod)
            else:
                if isinstance(version, str):
                    version = semantic_version.Version(version)

                center.installed.pin(self._mod, version)

            api.save_settings()
        else:
            raise NotImplemented

    def open_ver_edit(self, mod):
        ModVersionsWindow(mod)

    def apply_mod_ini_setting(self, state):
        settings = center.settings['mod_settings']
        if self._mod.mid not in settings:
            settings[self._mod.mid] = {}

        settings = settings[self._mod.mid]
        settings['parse_mod_ini'] = state == QtCore.Qt.Checked
        api.save_settings()

    def check_files(self):
        self.win.setCursor(QtCore.Qt.BusyCursor)

        task = CheckFilesTask(self._mod)
        task.done.connect(functools.partial(self.__check_files, task))
        run_task(task)

    def __check_files(self, task):
        report = ''

        for pkg, s, c, info in task.get_results():
            if pkg is None:
                if len(info['loose']) > 0:
                    report += '<h2>%s</h2>' % (self.tr('Loose files (%d)') % (len(info['loose'])))
                    report += '<ul><li>' + '</li><li>'.join(sorted(info['loose'])) + '</li></ul>'
            else:
                report += self.tr('<h2>%s (%d/%d files OK)</h2>') % (pkg.name, s, c)
                ok_count = len(info['ok'])
                corr_count = len(info['corrupt'])
                miss_count = len(info['missing'])

                report += '<ul>'

                if ok_count > 0:
                    if corr_count > 0 or miss_count > 0:
                        report += '<ul><li>%s<ul>' % self.tr('OK')
                        report += '<li>' + '</li><li>'.join(sorted(info['ok'])) + '</li>'
                        report += '</ul></li>'
                    else:
                        report += '<li>' + '</li><li>'.join(sorted(info['ok'])) + '</li>'

                if corr_count > 0:
                    report += '<li>%s<ul>' % self.tr('Corrupted')
                    report += '<li>' + '</li><li>'.join(sorted(info['corrupt'])) + '</li>'
                    report += '</ul></li>'

                if miss_count > 0:
                    report += '<li>%s<ul>' % self.tr('Missing')
                    report += '<li>' + '</li><li>'.join(sorted(info['missing'])) + '</li>'
                    report += '</ul></li>'

                report += '</ul>'

        self.win.logDisplay.setHtml(report)
        self.win.unsetCursor()

    def delete_loose_files(self):
        self.win.setCursor(QtCore.Qt.BusyCursor)

        task = CheckFilesTask(self._mod)
        task.done.connect(functools.partial(self.__delete_loose_files, task))
        run_task(task)

    # TODO: Should this be moved into a task?
    def __delete_loose_files(self, task):
        modpath = self._mod.folder

        for pkg, s, c, info in task.get_results():
            if pkg is None:
                for name in info['loose']:
                    item = os.path.join(modpath, name)
                    logging.info('Deleteing "%s"...', item)
                    os.unlink(item)

        self.win.unsetCursor()
        QtWidgets.QMessageBox.information(None, 'Knossos', self.tr('Done!'))

    def repair_files(self):
        try:
            run_task(InstallTask(self._mod.packages))
        except repo.ModNotFound as exc:
            QtWidgets.QMessageBox.critical(None, 'Knossos', self.tr("I can't repair this mod: %s") % str(exc))


class ModVersionsWindow(Window):
    _mod = None
    _versions = None

    def __init__(self, mod, window=True):
        super(ModVersionsWindow, self).__init__(window)

        self._mod = mod
        self._versions = []
        self._create_win(Ui_ModVersionsDialog, QDialog)
        self.win.setModal(True)

        self.label_tpl(self.win.label, MOD=mod.title)

        mods = list(center.mods.query_all(mod.mid))
        installed = list(center.installed.query_all(mod.mid))

        inst_versions = [m.version for m in installed]
        rem_versions = [m.version for m in mods]
        local_versions = set(inst_versions) - set(rem_versions)

        # Add all local-only versions
        for m in installed:
            if m.version in local_versions:
                mods.append(m)

        for m in mods:
            if len(m.packages) == 0:
                logging.warning('Version "%s" for mod "%s" (%s) is empty! (It has no packages!!)',
                    m.version, m.title, m.mid)
                continue

            label = str(m.version)
            if m.version in local_versions:
                label += self.tr(' (l)', 'This marks a version as *l*ocal-only. See also next line.')

            item = QtWidgets.QListWidgetItem(label, self.win.versionList)
            item.setData(QtCore.Qt.UserRole + 1, m)

            if m.version in local_versions:
                item.setToolTip(self.tr('This version is installed locally but not available anymore!'))

            if m.version in inst_versions:
                item.setCheckState(QtCore.Qt.Checked)
                item.setData(QtCore.Qt.UserRole, True)
            else:
                item.setCheckState(QtCore.Qt.Unchecked)
                item.setData(QtCore.Qt.UserRole, False)

            self._versions.append(item)

        self.win.applyButton.clicked.connect(self.apply_changes)
        self.win.cancelButton.clicked.connect(self.close)
        self.open()

    def apply_changes(self):
        install = set()
        uninstall = set()

        for item in self._versions:
            was = item.data(QtCore.Qt.UserRole)
            mod = item.data(QtCore.Qt.UserRole + 1)
            checked = item.checkState() == QtCore.Qt.Checked

            if was and not checked:
                uninstall |= set(center.installed.query(mod).packages)
            elif not was and checked:
                install |= set(mod.resolve_deps())

        install = center.mods.process_pkg_selection(install)

        if len(install) > 0:
            run_task(InstallTask(install, self._mod))

        if len(uninstall) > 0:
            run_task(UninstallTask(uninstall))

        self.close()


class ModInfoWindow(Window):
    _mod = None

    def __init__(self, mod, window=True):
        super(ModInfoWindow, self).__init__(window)

        self._mod = mod
        self._create_win(Ui_ModSettingsDialog, QDialog)

        self.win.modTitle.setText(mod.title)
        self.win.modLogo.setPixmap(QtGui.QPixmap(mod.logo))
        self.win.modDesc.setPlainText(mod.description)

        self.win.flagsTab.deleteLater()
        self.win.pkgTab.deleteLater()
        self.win.versionsTab.deleteLater()
        self.win.troubleTab.deleteLater()
        self.win.tabWidget.setCurrentIndex(0)

        self.open()


class LogViewer(Window):

    def __init__(self, path, window=True):
        super(LogViewer, self).__init__(window)

        self._create_win(Ui_LogDialog)
        self.win.pathLabel.setText(path)

        if not os.path.isfile(path):
            QtWidgets.QMessageBox.critical(None, 'Knossos',
                self.tr('Log file %s can\'t be shown because it\'s missing!') % path)
            return

        with open(path, 'r') as stream:
            self.win.content.setPlainText(stream.read())

        self.open()



## Copyright 2017 Knossos authors, see NOTICE file
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from __future__ import absolute_import, print_function

import sys
import os.path
import json
import logging
from collections import OrderedDict

from . import center, util, api, launcher, runner
from .tasks import run_task, CheckTask
from .qt import QtCore, QtWidgets, QtGui


tr = QtCore.QCoreApplication.translate


def get_settings():
    fso = {}

    fs2_bins = OrderedDict()
    fred_bins = OrderedDict()

    for mid, mvs in center.installed.mods.items():
        if mvs[0].mtype == 'engine':
            for v in mvs:
                for pkg in v.packages:
                    for exe in pkg.executables:
                        name = '%s - %s' % (v.title, exe['version'])
                        if exe.get('debug'):
                            name += ' (Debug)'

                        if exe.get('fred'):
                            fred_bins[name] = os.path.join(v.folder, exe['file'])
                        else:
                            fs2_bins[name] = os.path.join(v.folder, exe['file'])

    fso['fs2_bins'] = fs2_bins
    fso['fred_bins'] = fred_bins

    dev_info = get_deviceinfo()

    # ---Read fs2_open.ini or the registry---
    # Be careful with any change, the keys are all case sensitive.
    config = parse_fso_config()
    section = config.get('Default', {})

    # video settings
    if 'VideocardFs2open' in section:
        rawres = section['VideocardFs2open']

        try:
            res = rawres.split('(')[1].split(')')[0]
            res_width, res_height = res.split('x')
            depth = rawres.split(')x')[1][0:2]

            res_width = int(res_width)
            res_height = int(res_height)
        except (TypeError, IndexError):
            res_width = res_height = None
    else:
        res_width = None
        res_height = None
        depth = '32'

    texfilter = section.get('TextureFilter', '1')
    # af = section.get('OGL_AnisotropicFilter', None)
    # aa = section.get('OGL_AntiAliasSamples', None)

    # joysticks
    joystick_id = section.get('CurrentJoystick', None)

    # network settings
    net_connection = section.get('NetworkConnection', None)
    net_speed = section.get('ConnectionSpeed', None)
    net_port = section.get('ForcePort', None)

    net_ip = section.get('Network/CustomIP', None)

    section = config.get('Sound', {})

    # sound settings
    playback_device = section.get('PlaybackDevice', None)
    capture_device = section.get('CaptureDevice', None)
    enable_efx = section.get('EnableEFX', None)
    sample_rate = section.get('SampleRate', 44100)

    # ---Video settings---
    # Screen resolution
    modes = []
    if dev_info:
        for w, h in dev_info['modes']:
            if w > 800 and h > 600:
                modes.append('{0} x {1} ({2})'.format(w, h, get_ratio(w, h)))

    fso['modes'] = modes
    fso['depth'] = depth

    if res_width and res_height:
        fso['active_mode'] = '{0} x {1} ({2})'.format(res_width, res_height, get_ratio(res_width, res_height))
    elif len(modes) > 0:
        fso['active_mode'] = modes[0]
    else:
        fso['active_mode'] = ''

    try:
        index = int(texfilter)
    except TypeError:
        index = 0

    # If the SCP adds a new texture filder, we should change this part.
    if index > 1:
        index = 0
    fso['texfilter'] = index

    # ---Sound settings---
    if dev_info and dev_info['audio_devs']:
        snd_devs, snd_default, snd_captures, snd_default_capture = dev_info['audio_devs']

        if not playback_device and len(snd_devs) > 0:
            playback_device = snd_devs[0]

        if not capture_device and len(snd_captures) > 0:
            capture_device = snd_captures[0]

        fso['audio_devs'] = snd_devs
        fso['active_audio_dev'] = playback_device
        fso['default_audio_dev'] = snd_default

        fso['cap_devs'] = snd_captures
        fso['active_cap_dev'] = capture_device
        fso['default_cap_dev'] = snd_default_capture

        if util.is_number(sample_rate):
            sample_rate = int(sample_rate)
            if sample_rate > 0 and sample_rate < 1000000:
                fso['sample_rate'] = sample_rate
            else:
                fso['sample_rate'] = 44100
        else:
            fso['sample_rate'] = 44100

        # Fill EFX checkbox :
        fso['enable_efx'] = enable_efx == '1'

    # ---Joystick settings---
    fso['joysticks'] = dev_info['joysticks'] if dev_info else []
    
    # TODO: Implement UUID handling
    if util.is_number(joystick_id):
        if joystick_id == '99999':
            fso['joystick_id'] = 'No Joystick'
        else:
            fso['joystick_id'] = int(joystick_id)
    else:
        fso['joystick_id'] = 'No Joystick'

    # ---Network settings---
    net_connections_read = {'none': 0, 'dialup': 1, 'LAN': 2}
    if net_connection in net_connections_read:
        index = net_connections_read[net_connection]
    else:
        index = 2

    fso['net_type'] = index

    net_speeds_read = {'none': 0, 'Slow': 1, '56K': 2, 'ISDN': 3, 'Cable': 4, 'Fast': 5}
    if net_speed in net_speeds_read:
        index = net_speeds_read[net_speed]
    else:
        index = 5

    fso['net_speed'] = index
    fso['net_ip'] = net_ip
    fso['net_port'] = net_port

    kn_settings = center.settings.copy()
    del kn_settings['hash_cache']

    fso['has_voice'] = sys.platform == 'win32'

    return {
        'knossos': kn_settings,
        'languages': center.LANGUAGES,
        'has_log': launcher.log_path is not None,
        'fso': fso
    }


def save_fso_settings(new_settings):
    config = parse_fso_config()

    try:
        section = config.setdefault('Default', {})

        # Getting ready to write key=value pairs to the ini file
        # Set video
        new_res_width, new_res_height = new_settings['active_mode'].split(' (')[0].split(' x ')
        new_depth = new_settings['depth']

        section['VideocardFs2open'] = 'OGL -({0}x{1})x{2} bit'.format(new_res_width, new_res_height, new_depth)
        section['TextureFilter'] = new_settings['texfilter']

        # section['OGL_AntiAliasSamples'] = new_aa
        # section['OGL_AnisotropicFilter'] = new_af

        # sound
        section = config.setdefault('Sound', {})

        section['PlaybackDevice'] = new_settings.get('active_audio_dev', '')
        # ^ wxlauncher uses the same string as CaptureDevice, instead of what openal identifies as the playback device.
        # ^ So I do it the way openal is supposed to work, but I'm not sure FS2 really behaves that way

        section['CaptureDevice'] = new_settings.get('active_cap_dev', '')

        if new_settings['enable_efx']:
            section['EnableEFX'] = 1
        else:
            section['EnableEFX'] = 0

        section['SampleRate'] = new_settings['sample_rate']

        section = config['Default']

        # joystick
        if new_settings.get('joystick_id', 'No Joystick') == 'No Joystick':
            section['CurrentJoystick'] = 99999
        else:
            section['CurrentJoystick'] = new_settings['joystick_id']

        # networking
        # net_types = {0: 'none', 1: 'dialup', 2: 'LAN'}
        # new_net_connection = net_types[self._tabs['fso_network'].connectionType.currentIndex()]
        # section['NetworkConnection'] = new_net_connection

        # net_speeds = {0: 'none', 1: 'Slow', 2: '56K', 3: 'ISDN', 4: 'Cable', 5: 'Fast'}
        # new_net_speed = net_speeds[self._tabs['fso_network'].connectionSpeed.currentIndex()]
        # section['ConnectionSpeed'] = new_net_speed

        # new_net_ip = self._tabs['fso_network'].forceAddress.text()
        # if new_net_ip == '...':
        #     new_net_ip = ''

        # if new_net_ip == '':
        #     config.remove('Network/CustomIP')
        # else:
        #     config.setValue('Network/CustomIP'] = new_net_ip

        # new_net_port = self._tabs['fso_network'].localPort.text()
        # if new_net_port == '0':
        #     new_net_port = ''
        # elif new_net_port != '':
        #     try:
        #         new_net_port = int(new_net_port)
        #     except ValueError:
        #         new_net_port = ''

        # if new_net_port == '':
        #     config.remove(section + 'ForcePort')
        # else:
        #     section['ForcePort'] = new_net_port

        # Save the new configuration.
        write_fso_config(config)
    except:
        logging.exception('Failed to save the configuration!')
        QtWidgets.QMessageBox.critical(None, 'Knossos', tr('SettingsWindow', 'Failed to save the configuration!'))


def get_ratio(w, h):
    try:
        w = int(w)
        h = int(h)
    except TypeError:
        return '?'

    ratio = (1.0 * w / h)
    ratio = '{:.1f}'.format(ratio)
    if ratio == '1.3':
        ratio_string = '4:3'
    elif ratio == '1.6':
        ratio_string = '16:10'
    elif ratio == '1.8':
        ratio_string = '16:9'
    else:
        ratio_string = 'custom'
    return ratio_string


def parse_fso_config():
    config_file = os.path.join(api.get_fso_profile_path(), 'fs2_open.ini')
    if not os.path.isfile(config_file):
        return {}

    sections = {}
    current = None
    with open(config_file, 'r') as stream:
        for line in stream:
            if line.startswith('['):
                end = line.find(']')
                if not end:
                    logging.warning('Invalid line in INI file: %s' % line)
                    continue

                name = line[1:end]
                sections[name] = current = {}
            elif not line.startswith((';', '#')):
                middle = line.find('=')

                if not middle:
                    logging.warning('Invalid line in INI file: %s' % line)
                    continue

                key = line[:middle].strip()
                value = line[middle + 1:].strip()
                current[key] = value

    return sections


def write_fso_config(sections):
    config_file = os.path.join(api.get_fso_profile_path(), 'fs2_open.ini')

    with open(config_file, 'w') as stream:
        for name, pairs in sections.items():
            stream.write('[%s]\n' % name)

            for k, v in pairs.items():
                stream.write('%s = %s\n' % (k, v))

            stream.write('\n')


def get_deviceinfo():
    try:
        info = json.loads(util.check_output(launcher.get_cmd(['--deviceinfo'])).strip())
    except:
        logging.exception('Failed to retrieve device info!')

        QtWidgets.QMessageBox.critical(None, 'Knossos', tr('SettingsWindow',
            'There was an error trying to retrieve your device info (screen resolution, joysticks and audio' +
            ' devices). Please try again or report this error on the HLP thread.'))
        return None

    return info


def test_libs():
    sdl_path = center.settings['sdl2_path']
    oal_path = center.settings['openal_path']

    try:
        info = json.loads(util.check_output(launcher.get_cmd(['--lib-paths', sdl_path, oal_path])).strip())
    except:
        QtWidgets.QMessageBox.critical(None, 'Knossos',
            tr('SettingsWindow', 'An unknown error (a crash?) occurred while trying to load the libraries!'))
    else:
        if info['sdl2'] and info['openal']:
            msg = tr('SettingsWindow', 'Success! Both libraries were loaded successfully.')
        else:
            msg = tr('SettingsWindow', 'Error! One or both libraries failed to load.')

        msg += '\n\nSDL2: %s\nOpenAL: %s' % (
            info['sdl2'] if info['sdl2'] else tr('SettingsWindow', 'Not found'),
            info['openal'] if info['openal'] else tr('SettingsWindow', 'Not found')
        )
        QtWidgets.QMessageBox.information(None, 'Knossos', msg)


def save_setting(name, value):
    if name not in center.settings:
        raise Exception('Trying to set invalid setting "%s"!' % name)

    if name == 'max_downloads':
        if value != center.settings['max_downloads']:
            util.DL_POOL.set_capacity(value)
    elif name == 'language':
        # NOTE: This is deliberately not translated.
        QtWidgets.QMessageBox.information(None, 'Knossos', 'Please restart Knossos to complete the language change.')
    elif name == 'fs2_bin':
        path = os.path.join(center.settings['base_path'], 'bin', value)
        if not os.path.isfile(path):
            logging.warning('Tried to set a fs2_bin to a non-existing path: %s' % value)
            return

        old_bin = center.settings['fs2_bin']
        center.settings['fs2_bin'] = value

        if not api.get_fso_flags():
            # We failed to run FSO but why?
            rc = runner.run_fs2_silent(['-help'])
            if rc == -128:
                msg = tr('SettingsWindow', 'The FSO binary "%s" is missing!') % value
            elif rc == -127:
                # TODO: At this point we have run ldd twice already and the next call will run it again.
                # Is there any way to avoid this?
                _, missing = runner.fix_missing_libs(os.path.join(center.settings['base_path'], 'bin', value))
                msg = tr('SettingsWindow', 'The FSO binary "%s" is missing %s!') % (value, util.human_list(missing))
            else:
                msg = tr('SettingsWindow', 'The FSO binary quit with code %d!') % rc

            center.settings['fs2_bin'] = old_bin
            QtWidgets.QMessageBox.critical(None, 'Knossos', msg)
            return
        else:
            center.signals.fs2_bin_changed.emit()
    elif name == 'fred_bin':
        path = os.path.join(center.settings['base_path'], value)
        if not os.path.isfile(path):
            logging.warning('Tried to set a fred_bin to a non-existing path: %s' % value)
            return
    elif name == 'use_raven':
        if value:
            api.enable_raven()
        else:
            api.disable_raven()

    center.settings[name] = value
    api.save_settings()


def get_fso_log(self):
    logpath = os.path.join(api.get_fso_profile_path(), 'data/fs2_open.log')

    if not os.path.isfile(logpath):
        QtWidgets.QMessageBox.warning(None, 'Knossos',
            tr('SettingsWindow', 'Sorry, but I can\'t find the fs2_open.log file.\nDid you run the debug build?'))
    else:
        with open(logpath, 'r') as hdl:
            return logpath, hdl.read()


def get_knossos_log(self):
    with open(launcher.log_path, 'r') as hdl:
        return launcher.log_path, hdl.read()


def show_fso_cfg_folder(self):
    path = api.get_fso_profile_path()
    QtGui.QDesktopServices.openUrl(QtCore.QUrl('file://' + path))


def clear_hash_cache(self):
    util.HASH_CACHE = dict()
    run_task(CheckTask())
    QtWidgets.QMessageBox.information(None, 'Knossos', tr('SettingsWindow', 'Done!'))



## Copyright 2017 Knossos authors, see NOTICE file
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from __future__ import absolute_import, print_function

import sys
import os
import logging
import json
import shlex
import glob

from . import uhf
uhf(__name__)

from . import center, util, launcher, integration
from .qt import QtCore, QtWidgets
from .tasks import run_task, CheckUpdateTask, CheckTask, FetchTask, UninstallTask
from .windows import ModSettingsWindow, ModInstallWindow
from .repo import ModNotFound
from .ipc import IPCComm
from .runner import run_fs2_silent, run_mod

# TODO: Split this file up into smaller parts and move them into the respective modules
# (i.e. run_mod should be in runner).

ipc_block = None
translate = QtCore.QCoreApplication.translate


def save_settings():
    center.settings['hash_cache'] = dict()
    for path, info in util.HASH_CACHE.items():
        # Skip deleted files
        if os.path.exists(path):
            center.settings['hash_cache'][path] = info

    for mod in center.settings['cmdlines'].copy():
        if mod != '#default' and mod not in center.installed.mods:
            del center.settings['cmdlines'][mod]

    center.settings['pins'] = center.installed.pins

    with open(os.path.join(center.settings_path, 'settings.json'), 'w') as stream:
        json.dump(center.settings, stream)


def get_fso_flags():
    global fso_flags

    if center.settings['fs2_bin'] is None:
        return None

    if center.fso_flags is not None and center.fso_flags[0] == center.settings['fs2_bin']:
        return center.fso_flags[1]

    fs2_bin = os.path.join(center.settings['base_path'], 'bin', center.settings['fs2_bin'])
    if not os.path.isfile(fs2_bin):
        return None

    flags_path = os.path.join(center.settings['base_path'], 'flags.lch')
    rc = run_fs2_silent(['-get_flags', '-parse_cmdline_only'])

    flags = None

    if rc != 1 and rc != 0:
        logging.error('Failed to run FSO! (Exit code was %d)', rc)
    elif not os.path.isfile(flags_path):
        logging.error('Could not find the flags file "%s"!', flags_path)
    else:
        with open(flags_path, 'rb') as stream:
            flags = util.FlagsReader(stream)

    center.fso_flags = (center.settings['fs2_bin'], flags)
    return flags


def get_executables():
    exes = []

    for mid, mvs in center.installed.mods.items():
        for mod in mvs:
            for pkg in mod.packages:
                for item in pkg.executables:
                    name = mod.title + ' ' + item.get('version', '')
                    if item.get('debug', False):
                        name += ' (DEBUG)'

                    path = os.path.join(mod.folder, item['file'])
                    exes.append((name, path))

    bin_path = center.settings['base_path']
    if bin_path is not None:
        bin_path = os.path.abspath(os.path.join(bin_path, 'bin/custom'))
        if os.path.isdir(bin_path):
            if sys.platform == 'darwin':
                for app in glob.glob(os.path.join(bin_path, '*.app')):
                    name = os.path.basename(app)
                    exes.append((name, os.path.join(name, 'Contents', 'MacOS', name[:-4])))
            else:
                bins = glob.glob(os.path.join(bin_path, 'f*2_open_*'))

                for path in bins:
                    path = os.path.basename(path)

                    if not path.endswith(('.map', '.pdb')):
                        exes.append((path, path))

    return exes


def fetch_list():
    return run_task(FetchTask())


def get_cmdline(mod):
    if mod is None:
        return center.settings['cmdlines'].get('#default', [])[:]

    if mod.mid in center.settings['cmdlines']:
        return center.settings['cmdlines'][mod.mid][:]
    elif mod.cmdline != '':
        return shlex.split(mod.cmdline)
    else:
        return center.settings['cmdlines'].get('#default', [])[:]


def get_old_fso_profile_path():
    if sys.platform.startswith('linux'):
        leg_path = os.path.expanduser('~/.fs2_open')
    elif sys.platform == 'darwin':
        leg_path = os.path.expanduser('~/Library/FS2_Open')
    else:
        # TODO: This obviously won't work in most cases because we most likely won't be running FSO
        # from the base directory.
        leg_path = center.settings['base_path']

    return leg_path


_new_profile_path = None
def get_new_fso_profile_path():
    global _new_profile_path

    if _new_profile_path is None:
        try:
            _new_profile_path = util.check_output(launcher.get_cmd(['--fso-config-path'])).strip()
        except:
            logging.exception('Failed to retrieve FSO profile path from SDL!')
            _new_profile_path = 'None'

    return _new_profile_path


def get_fso_profile_path():
    path = get_new_fso_profile_path()
    leg_path = get_old_fso_profile_path()

    if leg_path and (path in ('', 'None') or
        (not os.path.exists(os.path.join(path, 'fs2_open.ini')) and os.path.exists(leg_path))
    ):
        profile_path = leg_path
    else:
        profile_path = path

    logging.info('Using profile path "%s".', profile_path)
    return profile_path


def read_fso_cmdline():
    # Look for the cmdline path.
    path = get_fso_profile_path()
    if path is None:
        return []

    path = os.path.join(path, 'data/cmdline_fso.cfg')

    # Read the current cmdline.
    cmdline = []
    if os.path.exists(path):
        try:
            with open(path, 'r') as stream:
                cmdline = shlex.split(stream.read().strip())
        except:
            logging.exception('Failed to read "%s", assuming empty cmdline.', path)

    for i, part in enumerate(cmdline):
        if part.strip() == '-mod':
            del cmdline[i]

            if len(cmdline) > i:
                del cmdline[i]

            break

    return cmdline


def check_retail_files():
    if center.settings['base_path'] is None:
        return

    has_retail = False
    fs2_path = os.path.join(center.settings['base_path'], 'FS2')
    if os.path.isdir(fs2_path):
        for item in os.listdir(fs2_path):
            if item.lower() == 'root_fs2.vp':
                has_retail = True
                break

    if has_retail:
        logging.debug('The FS2 path (%s) contains retail files!', center.settings['base_path'])
    else:
        logging.debug('The FS2 path (%s) does not contain retail files!', center.settings['base_path'])

    if has_retail != center.has_retail:
        center.has_retail = has_retail
        run_task(CheckTask())

    return has_retail


##############
# Public API #
##############


def get_mod(mid, version=None):
    if center.mods is None:
        QtWidgets.QMessageBox.critical(None, 'Knossos', translate('api', 'Hmm... I never got a mod list. Get a coder!'))
        return None
    else:
        try:
            return center.mods.query(mid, version)
        except ModNotFound:
            QtWidgets.QMessageBox.critical(None, 'Knossos', translate('api', 'Mod "%s" could not be found!') % mid)
            return None


# TODO: Create a proper window for this and move it to the windows module.
def uninstall_pkgs(pkgs, name=None, cb=None):
    titles = [pkg.name for pkg in pkgs if center.installed.is_installed(pkg)]

    if name is None:
        name = translate('api', 'these packages')

    msg = QtWidgets.QMessageBox()
    msg.setIcon(QtWidgets.QMessageBox.Question)
    msg.setText(translate('api', 'Do you really want to uninstall %s?') % name)
    msg.setInformativeText(translate('api', '%s will be removed.') % (', '.join(titles)))
    msg.setStandardButtons(QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
    msg.setDefaultButton(QtWidgets.QMessageBox.No)

    if msg.exec_() == QtWidgets.QMessageBox.Yes:
        task = UninstallTask(pkgs)
        if cb is not None:
            task.done.connect(cb)

        run_task(task)
        return True
    else:
        return False


#########
# Tools #
#########


def install_scheme_handler(interactive=True):
    logging.info('Installing scheme handler...')

    try:
        if integration.current.install_scheme_handler():
            if interactive:
                QtWidgets.QMessageBox.information(None, 'Knossos', translate('api', 'Done!'))
            return
    except:
        logging.exception('Failed to install the scheme handler!')

    QtWidgets.QMessageBox.critical(None, 'Knossos',
        translate('api', 'I probably failed to install the scheme handler.\nRun me as administrator and try again.'))


def setup_ipc():
    global ipc_block

    ipc_block = IPCComm(center.settings_path)
    ipc_block.messageReceived.connect(handle_ipc)
    ipc_block.listen()


def shutdown_ipc():
    global ipc_block

    if ipc_block:
        ipc_block.close()


def handle_ipc(msg):
    msg = msg.data().decode('utf8', 'ignore').strip()

    try:
        msg = json.loads(msg)
    except:
        logging.exception('Failed to parse IPC message %s.', msg)
        return

    if msg[0] == 'focus':
        center.main_win.win.activateWindow()
        center.main_win.win.raise_()
    elif msg[0] == 'run':
        if len(msg) < 2:
            QtWidgets.QMessageBox.critical(None, 'Knossos',
                translate('api.handle_ipc', 'The fso://run/<mod id> link is missing a parameter!'))
        else:
            mod = get_mod(msg[1])

            if mod is not None:
                run_mod(mod)
            else:
                QtWidgets.QMessageBox.critical(None, 'Knossos',
                    translate('api.handle_ipc', 'The mod "%s" was not found!') % msg[1])
    elif msg[0] == 'install':
        if len(msg) < 2:
            QtWidgets.QMessageBox.critical(None, 'Knossos',
                translate('api.handle_ipc', 'The fso://install/<mod id> link is missing a parameter!'))
            return

        mod = get_mod(msg[1])
        pkgs = []

        if not mod:
            # TODO: Maybe we should update the mod DB here?
            QtWidgets.QMessageBox.critical(None, 'Knossos',
                translate('api.handle_ipc', 'The mod "%s" was not found!') % msg[1])
            return

        if len(msg) > 2:
            for pname in msg[2:]:
                for pkg in mod.packages:
                    if pkg.name == pname:
                        pkgs.append(pkg)

        center.main_win.win.activateWindow()

        if mod.mid not in center.installed.mods:
            ModInstallWindow(mod, pkgs)
        else:
            QtWidgets.QMessageBox.information(None, 'Knossos',
                translate('api.handle_ipc', 'Mod "%s" is already installed!') % (mod.title))
    elif msg[0] == 'settings':
        center.main_win.win.activateWindow()

        if len(msg) == 1 or msg[1] == '':
            center.main_win.show_fso_settings()
        else:
            mod = get_mod(msg[1])

            if mod is None or mod.mid not in center.installed.mods:
                if mod is None:
                    name = msg[1]
                else:
                    name = mod.title

                QtWidgets.QMessageBox.information(None, 'Knossos',
                    translate('api.handle_ipc', 'Mod "%s" is not yet installed!') % (name))
            else:
                ModSettingsWindow(mod)
    else:
        QtWidgets.QMessageBox.critical(None, 'Knossos',
            translate('api.handle_ipc', 'The action "%s" is unknown!') % (msg[0]))


def _read_default_cmdline():
    if '#default' not in center.settings['cmdlines']:
        center.settings['cmdlines']['#default'] = read_fso_cmdline()


def enable_raven():
    try:
        from raven import Client
    except ImportError:
        logging.exception('Failed to import raven!')
        return False

    from raven.transport.threaded_requests import ThreadedRequestsHTTPTransport
    from raven.handlers.logging import SentryHandler

    if hasattr(sys, 'frozen'):
        from six.moves.urllib.parse import quote as urlquote
        center.SENTRY_DSN += '&ca_certs=' + urlquote(os.path.join(sys._MEIPASS, 'requests', 'cacert.pem'))

    center.raven = Client(
        center.SENTRY_DSN,
        release=center.VERSION,
        environment='debug' if center.DEBUG else 'production',
        transport=ThreadedRequestsHTTPTransport
    )
    center.raven.tags_context({
        'os': sys.platform
    })
    center.raven_handler = SentryHandler(center.raven, level=logging.ERROR)
    logging.getLogger().addHandler(center.raven_handler)

    return True


def disable_raven():
    if center.raven_handler:
        logging.getLogger().removeHandler(center.raven_handler)

    center.raven = None
    center.raven_handler = None


def init_self():
    setup_ipc()
    center.signals.fs2_path_changed.connect(_read_default_cmdline)
    center.signals.fs2_path_changed.connect(check_retail_files)

    if center.settings['update_notify'] and not center.VERSION.endswith('-dev'):
        run_task(CheckUpdateTask())



## Copyright 2017 Knossos authors, see NOTICE file
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from __future__ import absolute_import, print_function

import sys
import os.path
import logging
import re
import json
import semantic_version

from .qt import QtCore, QtGui, QtWidgets, QtWebChannel
from . import center, api, runner, repo, windows, tasks, util, settings

if not QtWebChannel:
    from .qt import QtWebKit


class WebBridge(QtCore.QObject):
    showWelcome = QtCore.Signal()
    showDetailsPage = QtCore.Signal('QVariant')
    showRetailPrompt = QtCore.Signal()
    updateModlist = QtCore.Signal('QVariantList', str)
    modProgress = QtCore.Signal(str, float, str)

    taskStarted = QtCore.Signal(float, str, list)
    taskProgress = QtCore.Signal(float, float, str)
    taskFinished = QtCore.Signal(float)

    def __init__(self, webView=None):
        super(WebBridge, self).__init__()

        if QtWebChannel:
            self.bridge = self
            page = webView.page()
            channel = QtWebChannel.QWebChannel(page)

            page.setWebChannel(channel)
            channel.registerObject('fs2mod', self)

            if center.DEBUG and os.path.isdir('../html') and os.environ.get('KN_BABEL') != 'True':
                link = os.path.abspath('../html/index.html')
                if sys.platform == 'win32':
                    link = '/' + link.replace('\\', '/')

                link = 'file://' + link
            else:
                link = 'qrc:///html/index.html'

            webView.load(QtCore.QUrl(link))

    @QtCore.Slot('QVariantList', result='QVariantMap')
    def finishInit(self, tr_keys):
        center.main_win.finish_init()

        trs = {}
        for k in tr_keys:
            trs[k] = QtCore.QCoreApplication.translate('modlist_ts', k)

        return trs

    @QtCore.Slot(str, str, result=str)
    def tr(self, context, msg):
        return QtCore.QCoreApplication.translate(context, msg)

    @QtCore.Slot(result=str)
    def getVersion(self):
        return center.VERSION

    @QtCore.Slot(result=bool)
    def isFsoInstalled(self):
        return api.is_fso_installed()

    @QtCore.Slot()
    def runGogInstaller(self):
        windows.GogExtractWindow()

    @QtCore.Slot(result='QVariantList')
    def getMods(self):
        return list(center.mods.get())

    @QtCore.Slot(result='QVariantList')
    def getInstalledMods(self):
        return list(center.installed.get())

    @QtCore.Slot(result='QVariantMap')
    def getUpdates(self):
        updates = center.installed.get_updates()
        result = {}
        for mid, items in updates.items():
            versions = result[mid] = {}
            for ver_a, ver_b in items.items():
                versions[str(ver_a)] = str(ver_b)

        return result

    @QtCore.Slot(str, str, result=bool)
    def isInstalled(self, mid, spec=None):
        if spec is None:
            return mid in center.installed.mods
        else:
            spec = util.Spec(spec)
            mod = center.installed.mods.get(mid, None)
            if mod is None:
                return False

            return spec.match(mod.version)

    @QtCore.Slot(str, str, result='QVariantMap')
    def query(self, mid, spec=None):
        if spec is not None:
            if spec == '':
                spec = None
            else:
                if re.search(r'^\d+', spec):
                    spec = '==' + spec

                try:
                    spec = util.Spec(spec)
                except:
                    logging.exception('Invalid spec "%s" passed to query()!', spec)
                    return -2

        try:
            return center.mods.query(mid, spec).get()
        except:
            return None

    @QtCore.Slot()
    def fetchModlist(self):
        api.fetch_list()

    @QtCore.Slot(bool, result='QVariantList')
    def requestModlist(self, async=False):
        if async:
            center.main_win.update_mod_list()
            return [None]
        else:
            return list(center.main_win.search_mods())

    @QtCore.Slot(str, str)
    def addRepo(self, repo_url, repo_name):
        repos = center.settings['repos']
        repos.append((repo_url, repo_name))

        api.save_settings()
        if center.main_win is not None:
            center.main_win.update_repo_list()

    @QtCore.Slot(result='QVariantList')
    def getRepos(self):
        return list(center.settings['repos'])

    @QtCore.Slot(str)
    def showTab(self, name):
        center.main_win.update_mod_buttons(name)

    @QtCore.Slot(str)
    def triggerSearch(self, term):
        center.main_win.perform_search(term)

    def _get_mod(self, mid, spec=None, mod_repo=None):
        if spec is not None:
            if spec == '':
                spec = None
            else:
                if re.search(r'^\d+', spec):
                    spec = '==' + spec

                try:
                    spec = util.Spec(spec)
                except:
                    logging.exception('Invalid spec "%s" passed to a web API function!', spec)
                    return -2

        if mod_repo is None:
            mod_repo = center.installed

        try:
            return mod_repo.query(mid, spec)
        except repo.ModNotFound:
            logging.exception('Couldn\'t find mod "%s" (%s)!', mid, spec)
            return -1

    @QtCore.Slot(str, str, result=int)
    def showAvailableDetails(self, mid, spec=None):
        mod = self._get_mod(mid, spec, center.mods)
        if mod in (-1, -2):
            return mod

        self.showDetailsPage.emit(mod.get())
        return 0

    @QtCore.Slot(str, str, result=int)
    def showInstalledDetails(self, mid, spec=None):
        mod = self._get_mod(mid, spec)
        if mod in (-1, -2):
            return mod

        self.showDetailsPage.emit(mod.get())
        return 0

    @QtCore.Slot(str, str, 'QStringList', result=int)
    def install(self, mid, spec=None, pkgs=None):
        mod = self._get_mod(mid, spec, center.mods)
        if mod in (-1, -2):
            logging.debug('fs2mod.install(%s, %s) = %d', mid, spec, mod)
            return mod

        if pkgs is None:
            pkgs = []

        if mod.parent == 'FS2' and not api.check_retail_files():
            self.showRetailPrompt.emit()
        else:
            windows.ModInstallWindow(mod, pkgs)

        return 0

    @QtCore.Slot(str, str, 'QStringList', result=int)
    def uninstall(self, mid, spec=None, pkgs=None):
        mod = self._get_mod(mid, spec)
        if mod in (-1, -2):
            return mod

        if len(pkgs) == 0:
            plist = mod.packages
        else:
            plist = []
            pfound = set()
            for pkg in mod.packages:
                if pkg.name in pkgs:
                    plist.append(pkg)
                    pfound.add(pkg.name)

            if len(pfound) < len(pkgs):
                # Some packages are missing
                pmissing = set(pkgs) - pfound
                logging.warning('Missing packages %s.', ', '.join(pmissing))
                return -2

        return api.uninstall_pkgs(plist, name=mod.title)

    @QtCore.Slot(str, str, result=int)
    def updateMod(self, mid, spec=None):
        mod = self._get_mod(mid, spec)

        if mod in (-1, -2):
            return mod

        all_vers = list(center.installed.query_all(mid))
        if len(all_vers) == 1:
            # Only one version is installed, let's update it.
            tasks.run_task(tasks.UpdateTask(mod))
        else:
            # Just install the new version
            cur_pkgs = list(mod.packages)
            for i, pkg in enumerate(cur_pkgs):
                cur_pkgs[i] = center.mods.query(mod.mid, None, pkg.name)

            tasks.run_task(tasks.InstallTask(cur_pkgs, cur_pkgs[0].get_mod()))

        center.main_win.update_mod_buttons('progress')
        return 0

    @QtCore.Slot(float)
    def abortTask(self, tid):
        if hasattr(center.main_win, 'abort_task'):
            center.main_win.abort_task(int(tid))

    @QtCore.Slot(str, str, result=int)
    def runMod(self, mid, spec=None):
        mod = self._get_mod(mid, spec)
        if mod in (-1, -2):
            return mod

        runner.run_mod(mod)
        return 0

    @QtCore.Slot(str, str, result=int)
    def runFredMod(self, mid, spec=None):
        mod = self._get_mod(mid, spec)
        if mod in (-1, -2):
            return mod

        runner.run_mod(mod, fred=True)
        return 0

    @QtCore.Slot(str, str, result=int)
    def showSettings(self, mid=None, spec=None):
        if mid is None or mid == '':
            windows.SettingsWindow()
            return 1
        else:
            mod = self._get_mod(mid, spec)
            if mod in (-1, -2):
                return mod

            windows.ModSettingsWindow(mod)
            return 1

    @QtCore.Slot(str, str, result=int)
    def showPackageList(self, mid=None, spec=None):
        mod = self._get_mod(mid, spec)
        if mod in (-1, -2):
            return mod

        win = windows.ModSettingsWindow(mod)
        win.show_pkg_tab()

        return 1

    @QtCore.Slot(str, str, result=int)
    def vercmp(self, a, b):
        try:
            a = semantic_version.Version(a)
            b = semantic_version.Version(b)
        except:
            # logging.exception('Someone passed an invalid version to vercmp()!')
            return 0

        return a.__cmp__(b)

    @QtCore.Slot(str)
    def openExternal(self, link):
        QtGui.QDesktopServices.openUrl(QtCore.QUrl(link))

    @QtCore.Slot(str, str, result=str)
    def browseFolder(self, title, path):
        return QtWidgets.QFileDialog.getExistingDirectory(None, title, path)

    @QtCore.Slot(str)
    def setBasePath(self, path):
        if not os.path.isdir(path):
            QtWidgets.QMessageBox.critical(None, 'Knossos', self.tr('WebBridge', 'The selected path is not a directory!'))
        else:
            center.settings['base_path'] = os.path.abspath(path)
            api.save_settings()
            tasks.run_task(tasks.FetchTask())
            center.main_win.check_fso()

    @QtCore.Slot(result=str)
    def getSettings(self):
        return json.dumps(settings.get_settings())

    @QtCore.Slot(str, str)
    def saveSetting(self, key, value):
        try:
            value = json.loads(value)
        except:
            logging.exception('Failed to decode new value for setting "%s"! (%s)' % (key, value))
        else:
            settings.save_setting(key, value)

    @QtCore.Slot(str)
    def saveFsoSettings(self, data):
        try:
            data = json.loads(data)
        except:
            logging.exception('Failed to decode new FSO settings! (%s)' % data)
        else:
            settings.save_fso_settings(data)

    @QtCore.Slot(result=str)
    def getDefaultFsoCaps(self):
        try:
            flags = api.get_fso_flags()
            if flags:
                flags = flags.to_dict()

            return json.dumps(flags)
        except:
            logging.exception('Failed to encode FSO flags!')

    @QtCore.Slot(result=str)
    def searchRetailData(self):
        for path in [r'C:\GOG Games\Freespace2']:
            if os.path.isdir(path):
                return path

        return ''

    @QtCore.Slot(str, result=bool)
    def copyRetailData(self, path):
        if os.path.isdir(path):
            tasks.run_task(tasks.GOGCopyTask(path, os.path.join(center.settings['base_path'], 'FS2')))
            return True
        elif os.path.isfile(path) and path.endswith('.exe'):
            tasks.run_task(tasks.GOGExtractTask(path, os.path.join(center.settings['base_path'], 'FS2')))
            return True
        else:
            QtWidgets.QMessageBox.critical(None, 'Knossos', self.tr('WebBridge', 'The selected path is not a directory!'))
            return False


if QtWebChannel:
    BrowserCtrl = WebBridge
else:
    class BrowserCtrl(object):
        _view = None
        _nam = None
        bridge = None

        def __init__(self, webView):
            self._view = webView
            self.bridge = WebBridge()

            settings = webView.settings()
            settings.setAttribute(QtWebKit.QWebSettings.DeveloperExtrasEnabled, True)

            frame = webView.page().mainFrame()
            frame.javaScriptWindowObjectCleared.connect(self.insert_bridge)

            link = 'qrc:///html/index.html'
            webView.load(QtCore.QUrl(link))

        def insert_bridge(self):
            frame = self._view.page().mainFrame()

            del self.bridge
            self.bridge = WebBridge()
            frame.addToJavaScriptWindowObject('fs2mod', self.bridge)
